from dataclasses import dataclass
from decimal import Decimal
from typing import Callable, Self, Optional

from sympy import Expr, Rational, Integer, Mul, Pow, S
from sympy.core.numbers import Half, Float


@dataclass
class Calculation:
    operation: Callable
    value0: Self | Decimal
    value1: Self | Decimal

    def result(self):
        value0: Calculation | Decimal = self.value0
        value1: Calculation | Decimal = self.value1

        if isinstance(value0, Calculation):
            value0 = value0.result()
        if isinstance(value1, Calculation):
            value1 = value1.result()

        return self.operation(value0, value1)


def expr_to_calc(expression: S) -> Calculation:
    calculation: Calculation | None = None
    operation: Optional[Callable] = None
    value0: Calculation | Decimal | None = None
    value1: Calculation | Decimal | None = None

    if len(expression.args) == 0:
        if isinstance(expression, Rational | Integer | Half):
            operation = Decimal.__truediv__
            value0 = Decimal(expression.p)
            value1 = Decimal(expression.q)
        elif isinstance(expression, Float):
            operation = Decimal.__add__
            value0 = Decimal(float(expression.num))
            value1 = Decimal(0)
        else:
            raise TypeError(f"Expression type '{type(expression)}' not yet supported.")
    else:
        if isinstance(expression, Mul):
            operation = Decimal.__mul__
        elif isinstance(expression, Pow):
            operation = Decimal.__pow__
        elif isinstance(expression, Expr):
            if isinstance(expression.args[0], Expr):
                calculation = expr_to_calc(expression=expression.args[0])
            elif isinstance(expression.args[0], int):
                calculation = Calculation(Decimal.__add__, value0=expression.args[0], value1=Decimal(0))
        else:
            raise TypeError(f"Expression type '{type(expression)}' not yet supported.")

        if calculation is None:
            value0 = expr_to_calc(expression=expression.args[0])
            value1 = expr_to_calc(expression=expression.args[1])

    if calculation is None and operation is not None and value0 is not None and value1 is not None:
        calculation = Calculation(operation=operation, value0=value0, value1=value1)
    else:
        raise ValueError(f"Unexpected values of operation '{operation}', value0 '{value0}', and value1 '{value1}'.")

    return calculation



from decimal import Decimal
from fractions import Fraction
from typing import NamedTuple

# TODO: use Numeric to make parts of this more generic
Numeric = int | float | Decimal | Fraction

# TODO: Point1Df as Vector1 in Decimal
#       although that would just be a wrapper for Decimal which is unlikely to be helpful
Point1Df = float
Point2Df = tuple[float, float]
Point1Di = int
Point2Di = tuple[int, int]


class Range1Df(NamedTuple):
    minimum: float
    maximum: float

    # TODO: implement as __len__
    def size(self) -> float:
        # Measuring stretch between fence posts
        # TODO: confirm that '1.0' should not be added here, because it is measuring between, not at
        return self.maximum - self.minimum + 1.0
        # return self.maximum - self.minimum


class Range1Di(NamedTuple):
    minimum: int
    maximum: int

    # TODO: implement as __len__
    def size(self) -> int:
        # Measuring amount of fence posts
        # TODO: confirm that '1' should not be added here,
        #       as range should be understood as [x:y[ rather than [x:y]
        return self.maximum - self.minimum + 1


class Range2Df(NamedTuple):
    x_axis: Range1Df
    y_axis: Range1Df

    def area(self) -> float:
        return self.x_axis.size() * self.y_axis.size()


class Range2Di(NamedTuple):
    x_axis: Range1Di
    y_axis: Range1Di

    def area(self) -> float:
        return self.x_axis.size() * self.y_axis.size()


# TODO: improve naming with words specific to bucketed int values and tiny-mini float values
def clamp_svg_value_to_range(svg_image_point: Point1Df, svg_image_axis: Range1Df) -> Point1Df:
    value: Point1Df
    if svg_image_point < svg_image_axis.minimum:
        value = svg_image_axis.minimum
    elif svg_image_point > svg_image_axis.maximum:
        value = svg_image_axis.maximum
    else:
        value = svg_image_point
    return value


# TODO: improve naming with words specific to bucketed int values and tiny-mini float values
def clamp_image_value_to_range(coloring_image_point: Point1Di, coloring_image_axis: Range1Di) -> Point1Di:
    value: Point1Di
    if coloring_image_point < coloring_image_axis.minimum:
        value = coloring_image_axis.minimum
    elif coloring_image_point > coloring_image_axis.maximum:
        value = coloring_image_axis.maximum
    else:
        value = coloring_image_point
    return value


# TODO: make naming generic to not focus on
def get_coloring_image_fraction(svg_image_point: Point1Df,
                                svg_axis_range: Range1Df) -> float:
    clamped_svg_image_y_axis_point: Point1Df = clamp_svg_value_to_range(svg_image_point=svg_image_point,
                                                                        svg_image_axis=svg_axis_range)
    axis_scaler: float = 1.0 / svg_axis_range.size()
    coloring_image_scaled_y_axis_point: float = clamped_svg_image_y_axis_point * axis_scaler
    return coloring_image_scaled_y_axis_point



from decimal import Decimal
from random import normalvariate

Variability = Decimal


def normalize_value(value: Decimal, minimum: Decimal, maximum: Decimal) -> Decimal:
    """Scale value position between minimum and maximum to [0:1] bound"""

    # Scale value and maximum relative to zero to enable finding fractional value by straightforward division
    zero_scaled_value: Decimal = value - minimum
    zero_scaled_maximum: Decimal = maximum - minimum

    # Divide scaled values
    normalized_value: Decimal = zero_scaled_value / zero_scaled_maximum

    return normalized_value


def clamp_or_normalize_value(value: Decimal, minimum: Decimal, maximum: Decimal) -> Decimal:
    new_value: Decimal

    if value < minimum:
        new_value = Decimal("0.0")
    elif value > maximum:
        new_value = Decimal("1.0")
    else:
        new_value = normalize_value(value=value, minimum=minimum, maximum=maximum)

    return new_value


def get_variability_amount(variability: Variability,
                           coloring_image_axis_size: Decimal) -> Decimal:
    if variability == Decimal("0.0"):
        variability_amount: Decimal = variability
    else:
        # TODO: get scaled or truncated value rather than retrying until value generated that fits in desired range
        #       search term: Truncated normal distribution
        # https://scipy.github.io/devdocs/reference/generated/scipy.stats.truncnorm.html
        # https://github.com/jessemzhang/tn_test/blob/master/truncated_normal/truncated_normal.py
        # package name: truncnorm
        # package name: pydistributions
        variation_value: Decimal = Decimal('Infinity')
        while variation_value < -1.0 or variation_value > 1.0:
            variation_value: Decimal = Decimal(normalvariate(mu=0.0, sigma=0.25))

        # Maximum extent of axis considered around point
        variability_radius: Decimal = variability * coloring_image_axis_size
        # Specific extent selected with normal distribution value
        variability_amount: Decimal = variation_value * variability_radius

    return variability_amount



from src.math_tompy.special import Range1Df, Point1Df, clamp_svg_value_to_range


def test_clamp_value_to_range_in_range_unchanged_success():
    # Setup
    axis0: Range1Df = Range1Df(minimum=0.0, maximum=1.0)
    point0: Point1Df = Point1Df(0.5)

    # Execution
    point1: Point1Df = clamp_svg_value_to_range(svg_image_point=point0, svg_image_axis=axis0)

    # Validation
    assert point0 == point1


def test_clamp_value_to_range_lower_increased_success():
    # Setup
    axis0: Range1Df = Range1Df(minimum=1.0, maximum=5.0)
    point0: Point1Df = Point1Df(-2.5)

    # Execution
    point1: Point1Df = clamp_svg_value_to_range(svg_image_point=point0, svg_image_axis=axis0)

    # Validation
    assert axis0[0] == point1


def test_clamp_value_to_range_higher_reduced_success():
    # Setup
    axis0: Range1Df = Range1Df(minimum=-10.0, maximum=22.0)
    point0: Point1Df = Point1Df(45.9)

    # Execution
    point1: Point1Df = clamp_svg_value_to_range(svg_image_point=point0, svg_image_axis=axis0)

    # Validation
    assert axis0[1] == point1




from enum import Enum, auto
from math import inf
from random import normalvariate

Variability = float


class Rounding(Enum):
    ROUND = auto()
    CEILING = auto()
    FLOOR = auto()


def round_to_nearest_multiple(number, round_to, rounding: Rounding = Rounding.ROUND):
    if round_to <= 0:
        raise ValueError(f"Rounding factor can not be 0 or negative.")

    remainder = number % round_to

    if rounding == Rounding.ROUND:
        nearest = number + round_to - remainder if 2 * remainder >= round_to else number - remainder
    elif rounding == Rounding.CEILING:
        nearest = number + round_to - remainder
    elif rounding == Rounding.FLOOR:
        nearest = number - remainder
    else:
        raise ValueError(f"rounding is not a valid Rounding value: '{rounding}")

    return nearest


def get_variability_amount(variability: Variability,
                           coloring_image_axis_size: int) -> float:
    if variability == 0.0:
        variability_amount: float = variability
    else:
        # TODO: get scaled or truncated value rather than retrying until value generated that fits in desired range
        #       search term: Truncated normal distribution
        # https://scipy.github.io/devdocs/reference/generated/scipy.stats.truncnorm.html
        # https://github.com/jessemzhang/tn_test/blob/master/truncated_normal/truncated_normal.py
        # package name: truncnorm
        # package name: pydistributions
        variation_value: float = inf
        while variation_value < -1.0 or variation_value > 1.0:
            variation_value: float = normalvariate(mu=0.0, sigma=0.25)

        # Maximum extent of axis considered around point
        variability_radius: float = variability * coloring_image_axis_size
        # Specific extent selected with normal distribution value
        variability_amount: float = variation_value * variability_radius

    return variability_amount



from decimal import Decimal

from src.math_tompy.decimal import normalize_value


def test_normalize_value_zero_half_one_success():
    # Setup
    value0: Decimal = Decimal("0.5")
    minimum0: Decimal = Decimal("0")
    maximum0: Decimal = Decimal("1")
    normalized_value0: Decimal = Decimal("0.5")

    # Execution
    normalized_value1: Decimal = normalize_value(value=value0, minimum=minimum0, maximum=maximum0)

    # Validation
    assert normalized_value0 == normalized_value1


def test_normalize_value_minus_one_zero_one_success():
    # Setup
    value0: Decimal = Decimal("0")
    minimum0: Decimal = Decimal("-1")
    maximum0: Decimal = Decimal("1")
    normalized_value0: Decimal = Decimal("0.5")

    # Execution
    normalized_value1: Decimal = normalize_value(value=value0, minimum=minimum0, maximum=maximum0)

    # Validation
    assert normalized_value0 == normalized_value1


def test_normalize_value_minus_ten_ten_thirty_success():
    # Setup
    value0: Decimal = Decimal("10")
    minimum0: Decimal = Decimal("-10")
    maximum0: Decimal = Decimal("30")
    normalized_value0: Decimal = Decimal("0.5")

    # Execution
    normalized_value1: Decimal = normalize_value(value=value0, minimum=minimum0, maximum=maximum0)

    # Validation
    assert normalized_value0 == normalized_value1
